@@ -1,12 +1,27 @@
+# file RODBC/R/RODBC.R
+# copyright (C) 1999-2007  M. Lapsley and B. D. Ripley
+#
+#  This program is free software; you can redistribute it and/or modify
+#  it under the terms of the GNU General Public License as published by
+#  the Free Software Foundation; either version 2 or 3 of the License
+#  (at your option).
+#
+#  This program is distributed in the hope that it will be useful,
+#  but WITHOUT ANY WARRANTY; without even the implied warranty of
+#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#  GNU General Public License for more details.
+#
+#  A copy of the GNU General Public License is available at
+#  http://www.r-project.org/Licenses/
+#
 #  Low level wrappers for odbc driver
 #
 #
 #
 .onLoad <- function(lib, pkg)
 {
-    library.dynam("RODBC", pkg, lib)
     if(is.null(getOption("dec")))
-        options(dec = Sys.localeconv()$decimal_point)
+        options(dec = Sys.localeconv()["decimal_point"])
 }
 
 .onAttach <- function(lib, pkg)
@@ -17,6 +32,7 @@
 .onUnload <- function(libpath)
 {
     odbcCloseAll()
+    .Call(C_RODBCTerm)
     library.dynam.unload("RODBC", libpath)
 }
 
@@ -24,9 +40,8 @@ odbcGetErrMsg <- function(channel)
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    err <- .Call("RODBCGetErrMsg", attr(channel, "handle_ptr"),
-                 PACKAGE = "RODBC")
-    .Call("RODBCClearError", attr(channel, "handle_ptr"), PACKAGE = "RODBC")
+    err <- .Call(C_RODBCGetErrMsg, attr(channel, "handle_ptr"))
+    .Call(C_RODBCClearError, attr(channel, "handle_ptr"))
     return(err)
 }
 
@@ -34,7 +49,7 @@ odbcClearError <- function(channel)
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    .Call("RODBCClearError", attr(channel, "handle_ptr"), PACKAGE = "RODBC")
+    .Call(C_RODBCClearError, attr(channel, "handle_ptr"))
     invisible()
 }
 
@@ -44,22 +59,41 @@ odbcReConnect <- function(channel, case, believeNRows)
         stop("Argument 'channel' must inherit from class RODBC")
     if(missing(case)) case <- attr(channel, "case")
     if(missing(believeNRows)) believeNRows <- attr(channel, "believeNRows")
-    odbcDriverConnect(attr(channel, "connection.string"), case, believeNRows)
+    odbcDriverConnect(attr(channel, "connection.string"), case, believeNRows,
+                      attr(channel, "rows_at_time"),
+                      colQuote = attr(channel, "colQuote"),
+                      tabQuote = attr(channel, "tabQuote"))
 }
 
 odbcConnect <-
-    function (dsn, uid = "", pwd = "", case = "nochange",
-              believeNRows = TRUE)
+    function (dsn, uid = "", pwd = "", ...)
 {
     st <- paste("DSN=", dsn, sep="")
     if(nchar(uid)) st <- paste(st, ";UID=", uid, sep="")
     if(nchar(pwd)) st <- paste(st, ";PWD=", pwd, sep="")
-    odbcDriverConnect(st, case = case, believeNRows = believeNRows)
+    odbcDriverConnect(st, ...)
 }
 
 odbcDriverConnect <-
-    function (connection = "", case = "nochange", believeNRows = TRUE)
+    function (connection = "", case = "nochange", believeNRows = TRUE,
+              colQuote, tabQuote = colQuote, DBMSencoding= "",
+              rows_at_time = 1000, bulk_add = NULL)
 {
+   id <- as.integer(1 + runif(1, 0, 1e5))
+   stat <- .Call(C_RODBCDriverConnect, as.character(connection), id,
+                 as.integer(believeNRows))
+   if(stat < 0) {
+       warning("ODBC connection failed")
+       return(stat)
+   }
+   res <- .Call(C_RODBCGetInfo, attr(stat, "handle_ptr"))
+   if(missing(colQuote)) colQuote <- ifelse(res[1] == "MySQL", "`", '"')
+   if(missing(case))
+       case <- switch(res[1],
+                      "MySQL" = "mysql",
+                      "PostgreSQL" = "postgresql",
+                      "ACCESS" = "msaccess",
+                      "nochange")
    switch(case,
 	toupper = case <- 1,
 	oracle = case <- 1,
@@ -70,24 +104,24 @@ odbcDriverConnect <-
 	mysql = case <- ifelse(.Platform$OS.type == "windows", 2, 0),
  	stop("Invalid case parameter: nochange | toupper | tolower | common db names")
 	)
-   id <- as.integer(1 + runif(1, 0, 1e5))
-   stat <- .Call("RODBCDriverConnect", as.character(connection), id,
-                 as.integer(believeNRows), PACKAGE = "RODBC")
-   if(stat < 0) {
-       warning("ODBC connection failed")
-       return(stat)
-   }
    case <- switch(case+1, "nochange", "toupper", "tolower")
-   return(structure(stat, class="RODBC", case=case, id = id,
-                    believeNRows=believeNRows))
+   if(is.null(bulk_add))
+       bulk_add <- .Call(C_RODBCCanAdd, attr(stat, "handle_ptr"))
+   structure(stat, class = "RODBC", case = case, id = id,
+             believeNRows = believeNRows, bulk_add = bulk_add,
+             colQuote = colQuote, tabQuote = tabQuote,
+             encoding = DBMSencoding,
+             rows_at_time = rows_at_time)
 }
 
-odbcQuery <- function(channel, query, rows_at_time = 1)
+odbcQuery <-
+    function(channel, query, rows_at_time = attr(channel, "rows_at_time"))
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    .Call("RODBCQuery", attr(channel, "handle_ptr"), as.character(query),
-          as.integer(rows_at_time), PACKAGE = "RODBC")
+    if(nchar(enc <- attr(channel, "encoding"))) query <- iconv(query, to=enc)
+    .Call(C_RODBCQuery, attr(channel, "handle_ptr"), as.character(query),
+          as.integer(rows_at_time))
 }
 
 odbcUpdate <-
@@ -96,52 +130,53 @@ odbcUpdate <-
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
+    if(nchar(enc <- attr(channel, "encoding"))) query <- iconv(query, to=enc)
     vflag <- 0
     if(verbose) vflag <- 1
     if(test) vflag <- 2
     ## apply the name mangling that was applied when the table was created
-    cnames<- gsub("[^A-Za-z0-9_]+", "", as.character(colnames(data)))
+    cnames <- mangleColNames(names(data))
     cnames <- switch(attr(channel, "case"),
                      nochange = cnames,
                      toupper = toupper(cnames),
                      tolower = tolower(cnames))
     for(i in 1:ncol(data))
-        if(!is.numeric(data[[i]]))
-           data[[i]] <- as.character(data[[i]])
-    .Call("RODBCUpdate", attr(channel, "handle_ptr"), as.character(query),
+        if(!is.numeric(data[[i]])) {
+            data[[i]] <- as.character(data[[i]])
+            if(nchar(enc)) data[[i]] <- iconv(data[[i]], to=enc)
+        }
+    .Call(C_RODBCUpdate, attr(channel, "handle_ptr"), as.character(query),
           data, cnames, as.integer(nrow(data)), as.integer(ncol(data)),
-          as.character(params), as.integer(vflag), PACKAGE = "RODBC")
+          as.character(params), as.integer(vflag))
 }
 
 odbcTables <- function(channel)
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    .Call("RODBCTables", attr(channel, "handle_ptr"), PACKAGE = "RODBC")
+    .Call(C_RODBCTables, attr(channel, "handle_ptr"))
 }
 
 odbcColumns <- function(channel, table)
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    .Call("RODBCColumns", attr(channel, "handle_ptr"), as.character(table),
-          PACKAGE = "RODBC")
+    .Call(C_RODBCColumns, attr(channel, "handle_ptr"), as.character(table))
 }
 
 odbcSpecialColumns <- function(channel, table)
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    .Call("RODBCSpecialColumns", attr(channel, "handle_ptr"), as.character(table),
-          PACKAGE = "RODBC")
+    .Call(C_RODBCSpecialColumns, attr(channel, "handle_ptr"),
+          as.character(table))
 }
 
 odbcPrimaryKeys <- function(channel, table)
 {
     if(!odbcValidChannel(channel))
         stop("first argument is not an open RODBC channel")
-    .Call("RODBCPrimaryKeys", attr(channel, "handle_ptr"), as.character(table),
-          PACKAGE = "RODBC")
+    .Call(C_RODBCPrimaryKeys, attr(channel, "handle_ptr"), as.character(table))
 }
 
 ## this does no error checking, but may add an error message
@@ -149,14 +184,14 @@ odbcColData <- function(channel)
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    .Call("RODBCColData", attr(channel, "handle_ptr"), PACKAGE = "RODBC")
+    .Call(C_RODBCColData, attr(channel, "handle_ptr"))
 }
 
 odbcNumCols <- function(channel)
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    .Call("RODBCNumCols", attr(channel, "handle_ptr"), PACKAGE = "RODBC")
+    .Call(C_RODBCNumCols, attr(channel, "handle_ptr"))
 }
 
 close.RODBC <- function(con, ...) odbcClose(con)
@@ -165,7 +200,7 @@ odbcClose <- function(channel)
 {
     if(!odbcValidChannel(channel))
        stop("argument is not an open RODBC channel")
-    res <- .Call("RODBCClose", attr(channel, "handle_ptr"), PACKAGE = "RODBC")
+    res <- .Call(C_RODBCClose, attr(channel, "handle_ptr"))
     if(res > 0) invisible(TRUE) else {
         warning(paste(odbcGetErrMsg(channel), sep="\n"))
         FALSE
@@ -174,7 +209,7 @@ odbcClose <- function(channel)
 
 odbcCloseAll <- function()
 {
-    .Call("RODBCCloseAll", PACKAGE = "RODBC")
+    .Call(C_RODBCCloseAll)
     invisible()
 }
 
@@ -184,8 +219,8 @@ odbcFetchRows <-
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    .Call("RODBCFetchRows", attr(channel, "handle_ptr"), max, buffsize,
-          as.character(nullstring), believeNRows, PACKAGE = "RODBC")
+    .Call(C_RODBCFetchRows, attr(channel, "handle_ptr"), max, buffsize,
+          as.character(nullstring), believeNRows)
 }
 
 odbcCaseFlag <- function (channel)
@@ -198,8 +233,8 @@ odbcCaseFlag <- function (channel)
 odbcGetInfo <- function(channel)
 {
     if(!odbcValidChannel(channel))
-       stop("first argument is not an open RODBC channel")
-    res <- .Call("RODBCGetInfo", attr(channel, "handle_ptr"), PACKAGE = "RODBC")
+       stop("argument is not an open RODBC channel")
+    res <- .Call(C_RODBCGetInfo, attr(channel, "handle_ptr"))
     names(res) <- c("DBMS_Name", "DBMS_Ver", "Driver_ODBC_Ver",
                     "Data_Source_Name", "Driver_Name", "Driver_Ver",
                     "ODBC_Ver", "Server_Name")
@@ -209,15 +244,14 @@ odbcGetInfo <- function(channel)
 odbcValidChannel <-  function(channel)
 {
     inherits(channel, "RODBC") && is.integer(channel) &&
-    .Call("RODBCcheckchannel", channel, attr(channel, "id"),
-          PACKAGE = "RODBC") > 0
+    .Call(C_RODBCcheckchannel, channel, attr(channel, "id")) > 0
 }
 
 odbcClearResults <-  function(channel)
 {
     if(!odbcValidChannel(channel))
        stop("first argument is not an open RODBC channel")
-    .Call("RODBCclearresults", attr(channel, "handle_ptr"), PACKAGE = "RODBC")
+    .Call(C_RODBCclearresults, attr(channel, "handle_ptr"))
     invisible()
 }
 
@@ -230,10 +264,45 @@ print.RODBC <- function(x, ...)
     invisible(x)
 }
 
-odbcSetAutoCommit <- function(channel, autoCommit=TRUE)
+odbcSetAutoCommit <- function(channel, autoCommit = TRUE)
 {
     if(!odbcValidChannel(channel))
          stop("first argument is not an open RODBC channel")
-    .Call("RODBCSetAutoCommit", attr(channel, "handle_ptr"), autoCommit,
-           PACKAGE = "RODBC")
+    .Call(C_RODBCSetAutoCommit, attr(channel, "handle_ptr"), autoCommit)
+}
+
+odbcEndTran <- function(channel, commit = TRUE)
+{
+    if(!odbcValidChannel(channel))
+         stop("first argument is not an open RODBC channel")
+    .Call(C_RODBCEndTran, attr(channel, "handle_ptr"), commit)
+}
+
+odbcBulkAdd <-
+    function(channel, query, data, params, first = 1, last = nrow(data),
+             test = FALSE, verbose = FALSE)
+{
+    if(!odbcValidChannel(channel))
+       stop("first argument is not an open RODBC channel")
+    vflag <- 0
+    if(verbose) vflag <- 1
+    if(test) vflag <- 2
+    ## apply the name mangling that was applied when the table was created
+    cnames <- mangleColNames(names(data))
+    cnames <- switch(attr(channel, "case"),
+                     nochange = cnames,
+                     toupper = toupper(cnames),
+                     tolower = tolower(cnames))
+    for(i in 1:ncol(data))
+        if(!is.numeric(data[[i]])) data[[i]] <- as.character(data[[i]])
+    .Call(C_RODBCAdd, attr(channel, "handle_ptr"), query, data, cnames,
+          as.integer(first), as.integer(last), as.integer(ncol(data)),
+          as.character(params), as.integer(vflag))
+}
+
+odbcDataSources <- function(type = c("all", "user", "system"))
+{
+    type <- match.arg(type)
+    type <- match(type, c("all", "user", "system"))
+    .Call(C_RODBCListDataSources, type)
 }
